@@ -55,6 +55,7 @@ class Searchable(object):
     _offset = 0
     _size = None
     _check_at_least = 0
+    _fromdoc = None
     _info = None
     _order_by = None
     _results = None
@@ -133,6 +134,9 @@ class Searchable(object):
         if check_at_least:
             body['check_at_least'] = check_at_least
 
+        if self._fromdoc is not None:
+            body['fromdoc'] = self._fromdoc
+
         if self._info is not None:
             body['info'] = self._info
 
@@ -243,6 +247,9 @@ class Searchable(object):
         """Get the offset of the first result item (0-based).
 
         """
+        if self._fromdoc is not None:
+            raise ValueError("result set offset is not meaningful when "
+                             "using fromdoc")
         return self._offset
 
     @property
@@ -370,9 +377,46 @@ class Searchable(object):
         size = self._size
         if size is not None and key >= size:
             raise IndexError("Index out of range of slice of results")
-        rank = self._offset + key
-        self._ensure_results_contain(rank)
-        return self._results.at_rank(rank)
+        if self._fromdoc is None:
+            rank = self._offset + key
+            self._ensure_results_contain(rank)
+            return self._results.at_rank(rank)
+        else:
+            assert self._offset == 0
+            self._ensure_results(0, size, self._check_at_least)
+            return self._results.items[key]
+
+    def fromdoc(self, doc_type, doc_id, offset = 0, size = 0,
+                fromdoc_pagesize = None):
+        """Get a subset of the result set, based on the position of a
+        particular base document.
+
+        :param doc_type: The type of the base document.
+        :param doc_id: The ID of the base document.
+        :param offset: The position offset for the start of the results to
+               return.  This may be negative to return results before the base
+               document; if this would result in trying to return results with
+               a rank less than zero, the calculated rank will be clipped to
+               zero.
+        :param size: The number of results to attempt to return.
+        :param fromdoc_pagesize: The number of results to calculate at a time
+               internally when working out what position the base document is
+               at.  This may usually be left as the server default, but can be
+               varied for performance reasons.
+
+        """
+        if self._offset != 0 or self._size is not None:
+            raise ValueError("fromdoc can not be used with a sliced result set")
+        result = TerminalQuery(self)
+        result._fromdoc = {
+            'type': doc_type,
+            'id': doc_id,
+            'from': offset,
+        }
+        if fromdoc_pagesize:
+            result._fromdoc['pagesize'] = fromdoc_pagesize
+        result._size = int(size)
+        return result
 
     def check_at_least(self, check_at_least):
         """Set the check_at_least value.
@@ -393,6 +437,20 @@ class Searchable(object):
     def order_by(self, field, ascending=None):
         """Set the sort order.
 
+        :param field: The name of a field, or self.RELEVANCE.
+        :param ascending: True if the sort order should be ascending (ie,
+               smallest values of the field get the highest weight).  If not
+               supplied, this will default to True when sorting by a field, and
+               False for RELEVANCE.
+
+        This may be called multiple times to order by multiple keys.
+        Alternatively, the order_by_multiple method may be used to do this.
+
+        In detail: if this is called when a sort order has already been set,
+        the previous sort order will be applied before the new one (ie, any
+        items which compare equal in the new order will be returned in the
+        order determined by the previously set sort order).
+
         """
         order_item = {}
         if field is self.RELEVANCE:
@@ -408,7 +466,26 @@ class Searchable(object):
             order_item['ascending'] = ascending
 
         result = TerminalQuery(self)
-        result._order_by = [order_item]
+        if result._order_by is None:
+            result._order_by = []
+        result._order_by.append(order_item)
+        return result
+
+    def order_by_multiple(self, orderings):
+        """Set the sort order, using multiple keys.
+
+        :param orderings: A sequence of ordering parameters, as supplied to the
+               order_by() method, in order of most significant first.
+
+        Any existing sort order is removed.
+
+        """
+        result = self
+        if result._order_by is not None:
+            result = TerminalQuery(result)
+            result._order_by = None
+        for order_item in orderings:
+            result = result.order_by(*order_item)
         return result
 
     @property
@@ -419,7 +496,33 @@ class Searchable(object):
         self._ensure_results_stats()
         return self._results.info
 
-    def calc_occur(self, prefix, doc_limit=None, result_limit=None,
+    def calc_facet_count(self, field, doc_limit=None, result_limit=None):
+        """Get facet counts for the given field in the matching documents.
+
+        Causes the search results to contain counts for each facet value seen
+        in the field, in decreasing order of occurrence.  The count entries are
+        of the form: [value, occurrence count].
+
+        :param slotname: The name or number of the the slot to read.
+        :param doc_limit: number of matching documents to stop checking after.  None=unlimited.  Integer or None.  Default=None
+        :param result_limit: number of terms to return results for.  None=unlimited.  Integer or None. Default=None
+
+        Note; all types being searched which contain the field must have been
+        configured to store facet values in the same slot.  The default
+        configuration will guarantee this, but if custom configuration results
+        in this constraint not being satisfied, an error will be returned.
+
+        """
+        result = TerminalQuery(self)
+        if result._info is None:
+            result._info = []
+        result._info.append({'facet_count': dict(field=field,
+                                                 doc_limit=doc_limit,
+                                                 result_limit=result_limit,
+                                                )})
+        return result
+
+    def calc_occur(self, group, prefix, doc_limit=None, result_limit=None,
                    get_termfreqs=False, stopwords=[]):
         """Get occurrence counts of terms in the matching documents.
 
@@ -430,17 +533,24 @@ class Searchable(object):
         [suffix, occurrence count] or [suffix, occurrence count, termfreq] if
         get_termfreqs was true.
 
+        :param group: group to check for terms in.
         :param prefix: prefix of terms to check occurrence for
         :param doc_limit: number of matching documents to stop checking after.  None=unlimited.  Integer or None.  Default=None
         :param result_limit: number of terms to return results for.  None=unlimited.  Integer or None. Default=None
         :param get_termfreqs: set to true to also get frequencies of terms in the db.  Boolean.  Default=False
         :param stopwords: list of stopwords - term suffixes to ignore.  Array of strings.  Default=[]
 
+        Note; if group is specified as an empty string, this can be used to
+        count occurrences of terms in all fields.  In this case, terms will be
+        represented by the group name, followed by a tab, followed by the
+        normal term.
+
         """
         result = TerminalQuery(self)
         if result._info is None:
             result._info = []
-        result._info.append({'occur': dict(prefix=prefix,
+        result._info.append({'occur': dict(group=group,
+                                           prefix=prefix,
                                            doc_limit=doc_limit,
                                            result_limit=result_limit,
                                            get_termfreqs=get_termfreqs,
@@ -448,7 +558,7 @@ class Searchable(object):
                                           )})
         return result
 
-    def calc_cooccur(self, prefix, doc_limit=None, result_limit=None,
+    def calc_cooccur(self, group, prefix, doc_limit=None, result_limit=None,
                      get_termfreqs=False, stopwords=[]):
         """Get cooccurrence counts of terms in the matching documents.
 
@@ -461,17 +571,24 @@ class Searchable(object):
         co-occurrence count, termfreq of suffix1, termfreq of suffix2] if
         get_termfreqs was true.
 
+        :param group: group to check for terms in.
         :param prefix: prefix of terms to check co-occurrence for
         :param doc_limit: number of matching documents to stop checking after.  None=unlimited.  Integer or None.  Default=None
         :param result_limit: number of terms to return results for.  None=unlimited.  Integer or None. Default=None
         :param get_termfreqs: set to true to also get frequencies of terms in the db.  Boolean.  Default=False
         :param stopwords: list of stopwords - term suffixes to ignore.  Array of strings.  Default=[]
 
+        Note; if group is specified as an empty string, this can be used to
+        count occurrences of terms in all fields.  In this case, terms will be
+        represented by the group name, followed by a tab, followed by the
+        normal term.
+
         """
         result = TerminalQuery(self)
         if result._info is None:
             result._info = []
-        result._info.append({'cooccur': dict(prefix=prefix,
+        result._info.append({'cooccur': dict(group=group,
+                                             prefix=prefix,
                                              doc_limit=doc_limit,
                                              result_limit=result_limit,
                                              get_termfreqs=get_termfreqs,
@@ -479,6 +596,16 @@ class Searchable(object):
                                             )})
         return result
 
+    if six.PY3:
+        __str__ = lambda x: x.__unicode__()
+    else:
+        __str__ = lambda x: unicode(x).encode('utf-8')
+
+    def __unicode__(self):
+        return six.u('<Query(%s)>') % (
+            self._build_search()
+        )
+
 
 class QueryIterator(object):
     """Iterate over the results of a query.
@@ -859,6 +986,7 @@ class TerminalQuery(Searchable):
         self._offset = orig._offset
         self._size = orig._size
         self._check_at_least = orig._check_at_least
+        self._fromdoc = copy.copy(orig._fromdoc)
         self._info = copy.copy(orig._info)
         self._order_by = copy.copy(orig._order_by)
         if slice is not None:
@@ -888,8 +1016,10 @@ class TerminalQuery(Searchable):
             stop = None
 
         # Set the starting offset.
-        oldstart = self._offset
-        self._offset = start + oldstart
+        if self._fromdoc is None:
+            self._offset = start + self._offset
+        else:
+            self._fromdoc['from'] = self._fromdoc['from'] + start
 
         # Update the size.
         oldsize = self._size
@@ -930,8 +1060,13 @@ class SearchResult(object):
         """
         self._object = object
 
-    def __str__(self):
-        return 'SearchResult(rank=%d, data=%s)' % (
+    if six.PY3:
+        __str__ = lambda x: x.__unicode__()
+    else:
+        __str__ = lambda x: unicode(x).encode('utf-8')
+
+    def __unicode__(self):
+        return six.u('SearchResult(rank=%d, data=%s)') % (
             self.rank, self.data,
         )
 
@@ -1059,19 +1194,24 @@ class SearchResults(object):
     def __getitem__(self, key):
         return self.items.__getitem__(key)
 
-    def __str__(self):
+    if six.PY3:
+        __str__ = lambda x: x.__unicode__()
+    else:
+        __str__ = lambda x: unicode(x).encode('utf-8')
+
+    def __unicode__(self):
         result = six.u('SearchResults(offset=%d, size_requested=%d, '
                        'check_at_least=%d, '
                        'matches_lower_bound=%d, '
                        'matches_estimated=%d, '
                        'matches_upper_bound=%d, '
-                       'items=[%s]' % (
+                       'items=[%s]') % (
             self.offset, self.size_requested, self.check_at_least,
             self.matches_lower_bound,
             self.matches_estimated,
             self.matches_upper_bound,
-            six.u(', '.join(str(item) for item in self.items)),
-        ))
+            ', '.join(str(item) for item in self.items),
+        )
         if self.info:
             result += six.u(', info=%s' % str(self.info))
         return result + six.u(')')
