@@ -55,7 +55,7 @@ class SearchTest(RestPoseTestCase):
 
     # Expected items for tests which return a single result
     expected_items_single = [
-        SearchResult(rank=0, data=expected_item_data),
+        SearchResult(rank=0, data=expected_item_data, results="dummy"),
     ]
 
     def check_results(self, results, offset=0, size_requested=None, check_at_least=0,
@@ -148,7 +148,14 @@ class SearchTest(RestPoseTestCase):
                                       't\\thello': {'positions': [1], 'wdf': 1},
                                       't\\tworld': {'positions': [2], 'wdf': 1},
                                       })
-        self.assertEqual(gotdoc.values, {})
+        self.assertEqual(gotdoc.values, {
+            '0': '\\x05Etest'
+                 '\\x04Fcat\\x06Fempty\\x04Ftag\\x05Ftest\\x05Ftext'
+                 '\\x06Mempty\\x04Ncat\\x04Ntag\\x05Ntest\\x05Ntext',
+            '1': '\\x05blurb',
+            '268435588': '\\x08greeting',
+            '268435592': '\\x05A tag',
+        })
 
     def test_getdoc(self):
         """Test that getting bits of a document works.
@@ -311,7 +318,7 @@ class SearchTest(RestPoseTestCase):
 
     def test_calc_cooccur(self):
         q = self.coll.doc_type("blurb").all()
-        results = self.coll.doc_type("blurb").search(q.calc_cooccur('t'))
+        results = self.coll.doc_type("blurb").search(q.calc_cooccur('t', ''))
         self.assertEqual(self.coll.status.get('doc_count'), 1)
         self.check_results(results, check_at_least=1,
                            items=self.expected_items_single,
@@ -319,32 +326,123 @@ class SearchTest(RestPoseTestCase):
                                'counts': [['hello', 'world', 1]],
                                'docs_seen': 1,
                                'terms_seen': 2,
-                               'prefix': 't',
+                               'group': 't',
+                               'prefix': '',
                                'type': 'cooccur'
                            }])
-        self.assertEqual(q.calc_cooccur('t').info,
+        self.assertEqual(q.calc_cooccur('t', '').info,
                          [{
                              'counts': [['hello', 'world', 1]],
                              'docs_seen': 1,
                              'terms_seen': 2,
-                             'prefix': 't',
+                             'group': 't',
+                             'prefix': '',
                              'type': 'cooccur'
                          }])
 
     def test_calc_occur(self):
         q = self.coll.doc_type("blurb").all()
-        results = self.coll.doc_type("blurb").search(q.calc_occur('t'))
+        results = self.coll.doc_type("blurb").search(q.calc_occur('t', ''))
         self.assertEqual(self.coll.status.get('doc_count'), 1)
         self.check_results(results, check_at_least=1,
                            items=self.expected_items_single,
                            info=[{
                                'counts': [['hello', 1], ['world', 1]],
                                'docs_seen': 1,
+                               'group': 't',
                                'terms_seen': 2,
-                               'prefix': 't',
+                               'prefix': '',
                                'type': 'occur'
                            }])
 
+    def test_tag_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('tag'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=1,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [['A tag', 1]],
+                               'docs_seen': 1,
+                               'fieldname': 'tag',
+                               'type': 'facet_count',
+                               'values_seen': 1,
+                           }])
+
+    def test_cat_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('cat'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=1,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [['greeting', 1]],
+                               'docs_seen': 1,
+                               'fieldname': 'cat',
+                               'type': 'facet_count',
+                               'values_seen': 1,
+                           }])
+
+    def test_type_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('type'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=1,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [['blurb', 1]],
+                               'docs_seen': 1,
+                               'fieldname': 'type',
+                               'type': 'facet_count',
+                               'values_seen': 1,
+                           }])
+
+    def test_text_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('text'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=0,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [],
+                               'docs_seen': 0, # Count of 0 because no stored
+                                               # facets for text fields
+                               'fieldname': 'text',
+                               'type': 'facet_count',
+                               'values_seen': 0,
+                           }])
+
+
+    def test_empty_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('empty'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=0,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [],
+                               'docs_seen': 0, # Count of 0 because this is a
+                                               # text field.
+                               'fieldname': 'empty',
+                               'type': 'facet_count',
+                               'values_seen': 0,
+                           }])
+
+    def test_missing_facet_count(self):
+        q = self.coll.doc_type("blurb").all()
+        results = self.coll.doc_type("blurb").search(q.calc_facet_count('missing'))
+        self.assertEqual(self.coll.status.get('doc_count'), 1)
+        self.check_results(results, check_at_least=0,
+                           items=self.expected_items_single,
+                           info=[{
+                               'counts': [],
+                               'docs_seen': 0, # Count of 0 because no field
+                                               # configuration for this field.
+                               'fieldname': 'missing',
+                               'type': 'facet_count',
+                               'values_seen': 0,
+                           }])
+
     def test_raw_query(self):
         results = self.coll.doc_type("blurb").search(dict(query=dict(matchall=True)))
 
@@ -450,10 +548,10 @@ class SearchTest(RestPoseTestCase):
 
         self.assertEqual(empty_type.search(empty_query[7:18]
                                            .check_at_least(3)
-                                           .calc_cooccur('t'))._raw,
+                                           .calc_cooccur('', ''))._raw,
                          missing_type.search(missing_query[7:18]
                                              .check_at_least(3)
-                                             .calc_cooccur('t'))._raw)
+                                             .calc_cooccur('', ''))._raw)
 
     def test_query_subscript(self):
         """Test subscript on a query.
@@ -563,7 +661,9 @@ class LargeSearchTest(RestPoseTestCase):
                                       '#\\tN': {},
                                       '#\\tNnum': {},
                                       })
-        self.assertEqual(gotdoc.values, { '268435599': '\\x02\\xb8\\xd0' })
+        self.assertEqual(gotdoc.values, { '0': '\\x04Fnum\\x04Nnum',
+                                          '1': '\\x03num',
+                                          '268435599': '\\x02\\xb8\\xd0' })
 
         gotdoc = self.coll.get_doc("other", "17")
         self.assertEqual(gotdoc.data, {
@@ -622,6 +722,49 @@ class LargeSearchTest(RestPoseTestCase):
         self.assertEqual(q1[242].data['num'], [0])
         self.assertRaises(IndexError, q1.__getitem__, 243)
 
+        # Test using multiple keys
+        q1 = q.order_by_multiple([['num', False]]) 
+        self.assertEqual(q1[0].data, self.make_doc(192))
+        self.assertEqual(q1[1].data, self.make_doc(191))
+        self.assertEqual(q1[242].data['num'], [0])
+        self.assertRaises(IndexError, q1.__getitem__, 243)
+
+        q1 = q.order_by_multiple([['num', True]]) 
+        self.assertEqual(q1[0].data['num'], [0])
+        self.assertEqual(q1[1].data['num'], [0])
+        self.assertEqual(q1[242].data, self.make_doc(192))
+        self.assertRaises(IndexError, q1.__getitem__, 243)
+
+        q1 = q.order_by_multiple([['type', False], ['num', True]]) 
+        self.assertEqual(q1[0].data['type'], ['other'])
+        self.assertEqual(q1[0].data['num'], [0])
+        self.assertEqual(q1[1].data['type'], ['other'])
+        self.assertEqual(q1[1].data['num'], [1])
+        self.assertEqual(q1[49].data['type'], ['other'])
+        self.assertEqual(q1[49].data['num'], [49])
+        self.assertEqual(q1[50].data['type'], ['num'])
+        self.assertEqual(q1[50].data['num'], [0])
+        self.assertEqual(q1[242].data, self.make_doc(192))
+        self.assertRaises(IndexError, q1.__getitem__, 243)
+
+        q1 = q.order_by_multiple([['type', True], ['num', True]]) 
+        self.assertEqual(q1[0].data, self.make_doc(0))
+        self.assertEqual(q1[1].data, self.make_doc(1))
+        self.assertEqual(q1[192].data, self.make_doc(192))
+        self.assertEqual(q1[193].data['type'], ['other'])
+        self.assertEqual(q1[193].data['num'], [0])
+        self.assertEqual(q1[242].data['type'], ['other'])
+        self.assertEqual(q1[242].data['num'], [49])
+        self.assertRaises(IndexError, q1.__getitem__, 243)
+
+        q1 = q.order_by_multiple([['num', True], ['type', True]]) 
+        self.assertEqual(q1[0].data, self.make_doc(0))
+        self.assertEqual(q1[1].data['type'], ['other'])
+        self.assertEqual(q1[1].data['num'], [0])
+        self.assertEqual(q1[2].data, self.make_doc(1))
+        self.assertEqual(q1[242].data, self.make_doc(192))
+        self.assertRaises(IndexError, q1.__getitem__, 243)
+
     def test_query_all(self):
         q = self.coll.doc_type("num").all().order_by('num')
 
@@ -701,3 +844,24 @@ class LargeSearchTest(RestPoseTestCase):
         self.assertEqual(q[:10].has_more, False)
         self.assertEqual(len(q), 9)
 
+    def test_fromdoc(self):
+        q = self.coll.doc_type('num').field.num.range(10, 100).order_by('num')
+        self.assertEqual(len(q), 91)
+        q1 = q.fromdoc('num', '60', -5, 10)
+        self.assertEqual(len(q1), 10)
+        self.assertEqual(q1[0].rank, 45)
+        self.assertEqual(q1[0].data['num'][0], 55)
+        q2 = q1[5:]
+        self.assertEqual(len(q2), 5)
+        self.assertEqual(q2[0].rank, 50)
+        self.assertEqual(q2[0].data['num'][0], 60)
+        q2 = q1[5:6]
+        self.assertEqual(len(q2), 1)
+        self.assertEqual(q2[0].rank, 50)
+        self.assertEqual(q2[0].data['num'][0], 60)
+        self.assertRaises(ValueError, q[:10].fromdoc, 'num', '60', -5, 10)
+        self.assertRaises(ValueError, q[10:].fromdoc, 'num', '60', -5, 10)
+        self.assertRaises(RequestFailed,
+                          q.fromdoc('num', '110', -5, 10).__getitem__, 0)
+        self.assertRaises(RequestFailed,
+                          q.fromdoc('num', '2000', -5, 10).__getitem__, 0)
